/* calmrisc16 simulator support code
   Copyright (C) 2000 Free Software Foundation, Inc.
   Contributed by Cygnus Solutions.

This file is part of the GNU simulators.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#define WANT_CPU
#define WANT_CPU_CALMRISC16BF

#include "sim-main.h"
#include "cgen-mem.h"
#include "cgen-ops.h"

/* The contents of BUF are in target byte order.  */
int
calmrisc16bf_fetch_register (SIM_CPU *current_cpu, int rn, unsigned char *buf, int len)
{
  if (rn <= GR_REGNUM_HI)
    SETTHI (buf, GET_H_GR (rn));
  else if (rn <= ER_REGNUM_HI)
    SETTHI (buf, GET_H_ER (rn - ER_REGNUM_LO + 8));
  else if (rn <= AR_REGNUM_HI)
    SETTSI (buf, GET_H_AR (rn - AR_REGNUM_LO + 8));
  else switch (rn)
    {
    case PC_REGNUM:
      SETTSI (buf, GET_H_PC ());
      break;
    case SPC_FIQ_REGNUM:
      SETTSI (buf,
	      (GET_H_SPR (H_SPR_SPCH_FIQ) << 16) | GET_H_SPR (H_SPR_SPCL_FIQ));
      break;
    case SPC_IRQ_REGNUM:
      SETTSI (buf,
	      (GET_H_SPR (H_SPR_SPCH_IRQ) << 16) | GET_H_SPR (H_SPR_SPCL_IRQ));
      break;
    case SR_REGNUM:
      SETTHI (buf, GET_H_SR ());
      break;
    case SSR_FIQ_REGNUM:
      SETTHI (buf, GET_H_SPR (H_SPR_SSR_FIQ));
      break;
    case SSR_IRQ_REGNUM:
      SETTHI (buf, GET_H_SPR (H_SPR_SSR_IRQ));
      break;
    case SSR_SWI_REGNUM:
      SETTHI (buf, GET_H_SPR (H_SPR_SSR_SWI));
      break;
    default:
      SETTSI (buf, 0xdeadbeef);
      return 0;
    }

  return -1;
}

/* The contents of BUF are in target byte order.  */
int
calmrisc16bf_store_register (SIM_CPU *current_cpu, int rn, unsigned char *buf, int len)
{
  if (rn <= GR_REGNUM_HI)
    SET_H_GR (rn, GETTHI (buf));
  else if (rn <= ER_REGNUM_HI)
    SET_H_ER (rn - ER_REGNUM_LO + 8, GETTHI (buf));
  else if (rn <= AR_REGNUM_HI)
    SET_H_AR (rn - AR_REGNUM_LO + 8, GETTSI (buf));
  else switch (rn)
    {
    case PC_REGNUM:
      SET_H_PC (GETTSI (buf));
      break;
    case SPC_FIQ_REGNUM:
      SET_H_SPR (H_SPR_SPCH_FIQ, (GETTSI (buf) >> 16) & 0xffff);
      SET_H_SPR (H_SPR_SPCL_FIQ, (GETTSI (buf)      ) & 0xffff);
      break;
    case SPC_IRQ_REGNUM:
      SET_H_SPR (H_SPR_SPCH_IRQ, (GETTSI (buf) >> 16) & 0xffff);
      SET_H_SPR (H_SPR_SPCL_IRQ, (GETTSI (buf)      ) & 0xffff);
      break;
    case SR_REGNUM:
      SET_H_SR (GETTHI (buf));
      break;
    case SSR_FIQ_REGNUM:
      SET_H_SPR (H_SPR_SSR_FIQ, GETTHI (buf));
      break;
    case SSR_IRQ_REGNUM:
      SET_H_SPR (H_SPR_SSR_IRQ, GETTHI (buf));
      break;
    case SSR_SWI_REGNUM:
      SET_H_SPR (H_SPR_SSR_SWI, GETTHI (buf));
      break;
    default:
      return 0;
    }

  return -1;
}

/* Cover fns to access the sr bits.  */
HI
calmrisc16bf_h_sr_get_handler (SIM_CPU *current_cpu)
{
  HI sr = (GET_H_T  () << 15) |
          (GET_H_PM () <<  6) |
          (GET_H_Z1 () <<  5) |
          (GET_H_Z0 () <<  4) |
          (GET_H_V  () <<  3) |
          (GET_H_TE () <<  2) |
          (GET_H_IE () <<  1) |
          (GET_H_FE ()      );
  return sr;
}

void
calmrisc16bf_h_sr_set_handler (SIM_CPU *current_cpu, HI newval)
{
  SET_H_T  ((newval >> 15) & 1);
  SET_H_Z1 ((newval >>  5) & 1);
  SET_H_Z0 ((newval >>  4) & 1);
  SET_H_V  ((newval >>  3) & 1);
  SET_H_TE ((newval >>  2) & 1);
  SET_H_IE ((newval >>  1) & 1);
  SET_H_FE ((newval      ) & 1);
  /* Some SR bits are restricted and can only be written when pm is already set.
     So set pm last here.  This gives the desired result when modifying the SR
     register both in user and privileged mode.  */
  SET_H_PM ((newval >>  6) & 1);
}

/* Semantic helper functions.  */

/* Pop the general registers represented by 'regmask' from the stack (where sp
   points).
*/
void
calmrisc16bf_pop_gr (SIM_CPU *current_cpu, PCADDR pc, SI regmask)
{
  /* The regmask is encoded as mmmmnnn where mmmm is 4 bits representing Rm
     and nnn is 3 bits representing Rn.  If Rn and Rm are the same register,
     then only pop it once.  */
  int Rm = (regmask >> 3) & 0xf;
  int Rn = (regmask & 0x7) | (Rm & 0x8);
  USI  sp = GET_H_AR (H_AR_SP);

  sp += 2;
  SET_H_GR (Rn, GETMEMHI (current_cpu, pc, sp));
  if (Rn != Rm)
    {
      sp += 2;
      SET_H_GR (Rm, GETMEMHI (current_cpu, pc, sp));
    }
  SET_H_AR (H_AR_SP, sp);
}

/* Pop the address registers represented by 'regmask' from the stack (where sp
   points).
*/
void
calmrisc16bf_pop_ar (SIM_CPU *current_cpu, PCADDR pc, SI regmask)
{
  /* The regmask is encoded as mmmnnn where mmm is 3 bits representing Am
     and nnn is 3 bits representing An.  If An and Am are the same register,
     then only pop it once.  */
  int Am = ((regmask >> 3) & 0x7) + 8;
  int An = (regmask & 0x7) + 8;
  USI  sp = GET_H_AR (H_AR_SP);

  /* Load halfwords at a time to avoid problems with forced alignment.  */
  sp += 2;
  SET_H_ER (An, GETMEMHI (current_cpu, pc, sp));
  sp += 2;
  SET_H_GR (An, GETMEMHI (current_cpu, pc, sp));
  if (An != Am)
    {
      sp += 2;
      SET_H_ER (Am, GETMEMHI (current_cpu, pc, sp));
      sp += 2;
      SET_H_GR (Am, GETMEMHI (current_cpu, pc, sp));
    }
  SET_H_AR (H_AR_SP, sp);
}

/* Push the general registers represented by 'regmask' to the stack (where sp
   points).
*/
void
calmrisc16bf_push_gr (SIM_CPU *current_cpu, PCADDR pc, SI regmask)
{
  /* The regmask is encoded as mmmmnnn where mmmm is 4 bits representing Rm
     and nnn is 3 bits representing Rn.  If Rn and Rm are the same register,
     then only pop it once.  */
  int Rm = (regmask >> 3) & 0xf;
  int Rn = (regmask & 0x7) | (Rm & 0x8);
  USI  sp = GET_H_AR (H_AR_SP);

  SETMEMHI (current_cpu, pc, sp, GET_H_GR (Rn));
  sp -= 2;
  if (Rn != Rm)
    {
      SETMEMHI (current_cpu, pc, sp, GET_H_GR (Rm));
      sp -= 2;
    }
  SET_H_AR (H_AR_SP, sp);
}

/* Push the address registers represented by 'regmask' to the stack (where sp
   points).
*/
void
calmrisc16bf_push_ar (SIM_CPU *current_cpu, PCADDR pc, SI regmask)
{
  /* The regmask is encoded as mmmnnn where mmm is 3 bits representing Am
     and nnn is 3 bits representing An.  If An and Am are the same register,
     then only pop it once.  */
  int Am = ((regmask >> 3) & 0x7) + 8;
  int An = (regmask & 0x7) + 8;
  USI  sp = GET_H_AR (H_AR_SP);

  /* Store halfwords at a time to avoid problems with forced alignment.  */
  SETMEMHI (current_cpu, pc, sp, GET_H_GR (An));
  sp -= 2;
  SETMEMHI (current_cpu, pc, sp, GET_H_ER (An));
  sp -= 2;
  if (An != Am)
    {
      SETMEMHI (current_cpu, pc, sp, GET_H_GR (Am));
      sp -= 2;
      SETMEMHI (current_cpu, pc, sp, GET_H_ER (Am));
      sp -= 2;
    }
  SET_H_AR (H_AR_SP, sp);
}

/* Cover fns to map Program and Data memory into a unified address space.  */
USI
calmrisc16bf_program_address (USI address)
{
  return CPU2INSN (address);
}

USI
calmrisc16bf_data_address (USI address)
{
  return CPU2DATA (address);
}

#if WITH_PROFILE_MODEL_P
/* Initialize cycle counting for an insn.
   FIRST_P is non-zero if this is the first insn in a set of parallel
   insns.  */

void
calmrisc16bf_model_insn_before (SIM_CPU *cpu, int first_p)
{
}

/* Record the cycles computed for an insn.
   LAST_P is non-zero if this is the last insn in a set of parallel insns,
   and we update the total cycle count.
   CYCLES is the cycle count of the insn.  */

void
calmrisc16bf_model_insn_after (SIM_CPU *cpu, int last_p, int cycles)
{
}

int
calmrisc16bf_model_calmrisc16_u_exec (SIM_CPU *cpu, const IDESC *idesc,
				      int unit_num, int referenced)
{
  int cycles = idesc->timing->units[unit_num].done;
  return cycles;
}

#endif /* WITH_PROFILE_MODEL_P */
