/* calmrisc16 exception, interrupt, and trap (EIT) support
   Copyright (C) 2000 Free Software Foundation, Inc.
   Contributed by Cygnus Solutions.

This file is part of the GNU simulators.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

#define WANT_CPU calmrisc16bf
#define WANT_CPU_CALMRISC16BF

#include "sim-main.h"
#include "targ-vals.h"
#include "cgen-engine.h"
#include "cgen-ops.h"

/* The semantic code invokes this for invalid (unrecognized) instructions.  */

SEM_PC
sim_engine_invalid_insn (SIM_CPU *current_cpu, IADDR cia, SEM_PC vpc)
{
  SIM_DESC sd = CPU_STATE (current_cpu);
  sim_engine_halt (sd, current_cpu, NULL, cia, sim_stopped, SIM_SIGILL);
  return vpc;
}

/* Process an address exception.  */

void
calmrisc16_core_signal (SIM_DESC sd, SIM_CPU *current_cpu, sim_cia cia,
		  unsigned int map, int nr_bytes, address_word addr,
		  transfer_type transfer, sim_core_signals sig)
{
  sim_core_signal (sd, current_cpu, cia, map, nr_bytes, addr,
		   transfer, sig);
}

/* Read/write functions for system call interface.  */

static int
syscall_read_mem (host_callback *cb, struct cb_syscall *sc,
		  unsigned long taddr, char *buf, int bytes)
{
  SIM_DESC sd = (SIM_DESC) sc->p1;
  SIM_CPU *cpu = (SIM_CPU *) sc->p2;

  return sim_core_read_buffer (sd, cpu, read_map, buf, taddr, bytes);
}

static int
syscall_write_mem (host_callback *cb, struct cb_syscall *sc,
		   unsigned long taddr, const char *buf, int bytes)
{
  SIM_DESC sd = (SIM_DESC) sc->p1;
  SIM_CPU *cpu = (SIM_CPU *) sc->p2;

  return sim_core_write_buffer (sd, cpu, write_map, buf, taddr, bytes);
}

/* Setup for a TRQ interrupt.  */
PCADDR
calmrisc16_setup_trq (SIM_CPU *current_cpu, PCADDR pc)
{
  SET_H_SPCH_IRQ ((pc >> 8) & 0xffff);
  SET_H_SPCL_IRQ (pc & 0xffff);
  SET_H_SSR_IRQ (GET_H_SR ());
  /* The pm, ie and te bits are restricted and can only be written when pm
     is already set.  So, set pm directly and before setting ie and te.  */
  CPU (h_pm) = 1;
  SET_H_IE (0);
  SET_H_TE (0);
  /* The TRQ trap vector is stored in program memory at address 0x000006.  */
  pc = GETMEMHI (current_cpu, pc, calmrisc16bf_program_address (0x000006));
  return pc;
}

/* Subroutine of calmrisc16_swi to setup for SWI.  */
static PCADDR
setup_swi (SIM_CPU *current_cpu, PCADDR pc, int imm6)
{
  SET_H_A14 (pc + 2);
  SET_H_SSR_SWI (GET_H_SR ());
  /* The pm, ie and te bits are restricted and can only be written when pm
     is already set.  So, set pm directly and before setting ie and te.  */
  CPU (h_pm) = 1;
  SET_H_IE (0);
  SET_H_TE (0);
  return imm6 << 2;
}

/* Trap support.
   The result is the pc address to continue at.
   Preprocessing like saving the various registers must be done here.  */
USI
calmrisc16_swi (SIM_CPU *current_cpu, PCADDR pc, int imm6)
{
  SIM_DESC sd = CPU_STATE (current_cpu);
  host_callback *cb = STATE_CALLBACK (sd);

  /* TRAP_SYSCALL and TRAP_BREAKPOINT have been selected as traps not available
     to the programmer, so honour them always. This allows us to execute
     syscalls and breakpoints and, most importantly, to stop execution, even if
     --environment=operating.  */
  switch (imm6)
    {
    case TRAP_SYSCALL :
      {
	CB_SYSCALL s;

	CB_SYSCALL_INIT (&s);
	s.func = GET_H_GR (0);
	s.arg1 = GET_H_AR (8);
	s.arg2 = GET_H_AR (9);
	s.arg3 = GET_H_AR (10);

	if (s.func == TARGET_SYS_exit)
	  {
	    sim_engine_halt (sd, current_cpu, NULL, pc, sim_exited, s.arg1);
	  }

	s.p1 = (PTR) sd;
	s.p2 = (PTR) current_cpu;
	s.read_mem = syscall_read_mem;
	s.write_mem = syscall_write_mem;
	cb_syscall (cb, &s);
	SET_H_GR (2, s.errcode);
	SET_H_AR (8, s.result);
	SET_H_AR (9, s.result2);
	break;
      }

    case TRAP_BREAKPOINT:
#ifdef SIM_HAVE_BREAKPOINTS
      /* Check for breakpoints "owned" by the simulator first, regardless
	 of --environment.  It will return if it can't process the
	 breakpoint.  */
      sim_handle_breakpoint (sd, current_cpu, pc);
#endif
      sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGTRAP);
      break;

    default :
      {
	USI new_pc = setup_swi (current_cpu, pc, imm6);
	return new_pc;
      }
    }

  /* Fake a "ret_swi" insn.  All we need to do is update pc.  */
  return pc + 2;
}

/* Handle the BREAK insn.  */
void
calmrisc16_break (SIM_CPU *current_cpu)
{
  IADDR pc;
  SIM_DESC sd = CPU_STATE (current_cpu);

#ifdef SIM_HAVE_BREAKPOINTS
  /* First try sim-break.c.  If it's a breakpoint the simulator "owns"
     it doesn't return.  Otherwise it returns and let's us try.  */
  pc = GET_H_PC ();
  sim_handle_breakpoint (sd, current_cpu, pc);
  /* Fall through.  */
#endif

  if (STATE_ENVIRONMENT (sd) != OPERATING_ENVIRONMENT)
    {
#ifndef SIM_HAVE_BREAKPOINTS
      pc = GET_H_PC ();
#endif
      sim_engine_halt (sd, current_cpu, NULL, pc, sim_stopped, SIM_SIGTRAP);
    }

  /* Otherwise, nothing to do.  */
  return;
}
