# r0, r5 are reserved
# a15 (and hence r15 and e15) is sp and is reserved

	.macro start
	.data
failmsg:
	.ascii "fail\n"
passmsg:
	.ascii "pass\n"
	.text
	.global _start
_start:
	ld	sp,0xffffe	; TODO -- what's a good value for this?
	.endm

; Exit with return code
	.macro exit rc
	ld	a8,\rc
	ld	r0,1
	swi	0
	.endm

; Pass the test case
	.macro pass
	ld	a10,5
	ld	a9,passmsg
	ld	a8,1
	ld	r0,5
	swi	0
	exit	0
	.endm

; Fail the test case
	.macro fail
	ld	a10,5
	ld	a9,failmsg
	ld	a8,1
	ld	r0,5
	swi	0
	exit	1
	.endm
;---------------------------------------------------
; Set a general register with an immediate value
	.macro set_gr_immed reg val
	ld	\reg,\val
	.endm

; Test a general register against an immediate
	.macro test_gr_immed reg val
	cmp	eq,\reg,\val
	brt	exit\@
	fail
exit\@:
	.endm

;---------------------------------------------------
; Set an address register with an immediate value
	.macro set_ar_immed reg val
	ld	\reg,\val
	.endm

; Decrement an address register with an immediate value
	.macro dec_ar_immed reg val
	sub	\reg,\val
	.endm

; Increment an address register with an immediate value
	.macro inc_ar_immed reg val
	add	\reg,\val
	.endm

; Test an address register against an immediate
	.macro test_ar_immed reg val
	cmp	eq,\reg,\val
	brt	exit\@
	fail
exit\@:
	.endm

;---------------------------------------------------
; Set an address register with another address register
	.macro set_ar_ar targ src
	ld	\targ,\src
	.endm

; Test an address register against another address register
	.macro test_ar_ar reg1 reg2
	cmp	eq,\reg1,\reg2
	brt	exit\@
	fail
exit\@:
	.endm

;---------------------------------------------------
; Set an extended register with an immediate value
	.macro set_er_immed reg val
	set_gr_immed	r5,\val
	ld		\reg,r5
	.endm

; Test an extended register against an immediate
	.macro test_er_immed reg val
	ld		r5,\reg
	test_gr_immed	r5,\val
	.endm

;---------------------------------------------------
; Set a word in memory with an immediate value
	.macro set_mem_immed addr val
	set_gr_immed	r5,\val
	ldw		@[\addr+0],r5
	.endm

; Test an extended register against an immediate
	.macro test_mem_immed addr val
	ldw		r5,@[\addr+0]
	test_gr_immed	r5,\val
	.endm

;---------------------------------------------------
; Test a special register against an immediate value
	.macro test_spr_immed reg val
	ld		r0,\reg
	test_gr_immed	r0,\val
	.endm

;---------------------------------------------------
; Set the t bit
	.macro set_t
	cmp	eq,r0,r0
	.endm

; Clear the t bit
	.macro clear_t
	cmp	gt,r0,r0
	.endm

; Test the t bit
	.macro test_t val
	.if (\val == 0)
	brf	exit\@
	.else
	brt	exit\@
	.endif
	fail
exit\@:
	.endm

; Set the z0 bit
	.macro set_z0
	setsr	z0
	.endm

; Set the z1 bit
	.macro set_z1
	setsr	z1
	.endm

; Set the v bit
	.macro set_v
	setsr	v
	.endm

; Set the fe bit
	.macro set_fe
	setsr	fe
	.endm

; Set the ie bit
	.macro set_ie
	setsr	ie
	.endm

; Set the te bit
	.macro set_te
	setsr	te
	.endm

; Set the pm bit
	.macro set_pm
	setsr	pm
	.endm

; Clear the z0 bit
	.macro clear_z0
	clrsr	z0
	.endm

; Clear the z1 bit
	.macro clear_z1
	clrsr	z1
	.endm

; Clear the v bit
	.macro clear_v
	clrsr	v
	.endm

; Clear the fe bit
	.macro clear_fe
	clrsr	fe
	.endm

; Clear the ie bit
	.macro clear_ie
	clrsr	ie
	.endm

; Clear the te bit
	.macro clear_te
	clrsr	te
	.endm

; Clear the pm bit
	.macro clear_pm
	clrsr	pm
	.endm

; Test the z0 bit
	.macro test_z0 val
	tstsr	z0
	test_t	!(\val)
	.endm

; Test the z1 bit
	.macro test_z1 val
	tstsr	z1
	test_t	!(\val)
	.endm

; Test the v bit
	.macro test_v val
	tstsr	v
	test_t	!(\val)
	.endm

; Test the fe bit
	.macro test_fe val
	tstsr	fe
	test_t	!(\val)
	.endm

; Test the ie bit
	.macro test_ie val
	tstsr	ie
	test_t	!(\val)
	.endm

; Test the te bit
	.macro test_te val
	tstsr	te
	test_t	!(\val)
	.endm

; Test the pm bit
	.macro test_pm val
	tstsr	pm
	test_t	!(\val)
	.endm

;---------------------------------------------------
; Branch to a nearby label
	.macro goto addr
	bra	\addr
	.endm
;---------------------------------------------------
; Jump to an address in an address register
	.macro jump addr
	jmp	\addr
	.endm

;---------------------------------------------------
; Setup a state consistent with a FIQ handler
	.macro setup_fiq handler,return
	ld	r0,%hi(\return)
	ld	spch_fiq,r0
	ld	r0,%lo(\return)
	ld	spcl_fiq,r0
	ld	r0,sr
	ld	ssr_fiq,r0
	and	r0,0xfff8	; no effect is pm is off
	or	r0,0x0040	; no effect. Can't set pm if it is off
	ld	sr,r0
	bra	\handler	
	.endm

;---------------------------------------------------
; Setup a state consistent with an IRQ handler
	.macro setup_irq handler,return
	ld	r0,%hi(\return)
	ld	spch_irq,r0
	ld	r0,%lo(\return)
	ld	spcl_irq,r0
	ld	r0,sr
	ld	ssr_irq,r0
	and	r0,0xfff9	; no effect is pm is off
	or	r0,0x0040	; no effect. Can't set pm if it is off
	ld	sr,r0
	bra	\handler	
	.endm

;---------------------------------------------------
; Setup a state consistent with a SWI handler
	.macro setup_swi handler,return
	ld	a14,\return
	ld	r0,sr
	ld	ssr_swi,r0
	and	r0,0xfff9	; no effect is pm is off
	or	r0,0x0040	; no effect. Can't set pm if it is off
	ld	sr,r0
	bra	\handler	
	.endm
